#import "KinectManager.h"

// IMPORTANT: REMEMBER TO CHANGE YOUR MASTER DEVICE LIST AFTER BUILDING THIS BUNDLE

#define GESTURE_TO_USE "Click" 

typedef enum
{
	IN_SESSION,
	NOT_IN_SESSION,
	QUICK_REFOCUS
} SessionState;


xn::Context g_Context;
xn::ScriptNode g_ScriptNode;
xn::DepthGenerator g_DepthGenerator;
xn::HandsGenerator g_HandsGenerator;
xn::HandsGenerator g_HandsGenerator2;
xn::GestureGenerator g_GestureGenerator;

// NITE objects
XnVSessionManager* g_pSessionManager;

SessionState g_SessionState = NOT_IN_SESSION;


#define CHECK_RC(rc, what)											\
if (rc != XN_STATUS_OK)											\
{																\
printf("%s failed: %s\n", what, xnGetStatusString(rc));		\
return;													\
}

#define CHECK_ERRORS(rc, errors, what)		\
if (rc == XN_STATUS_NO_NODE_PRESENT)	\
{										\
XnChar strError[1024];				\
errors.ToString(strError, 1024);	\
printf("%s\n", strError);			\
return;						\
}

void XN_CALLBACK_TYPE SessionProgress(const XnChar* strFocus, const XnPoint3D& ptFocusPoint, XnFloat fProgress, void* UserCxt) {
	printf("Session progress (%6.2f,%6.2f,%6.2f) - %6.2f [%s]\n", ptFocusPoint.X, ptFocusPoint.Y, ptFocusPoint.Z, fProgress,  strFocus);
}
// callback for session start
void XN_CALLBACK_TYPE SessionStart(const XnPoint3D& ptFocusPoint, void* UserCxt) {
	KinectManager *me = (KinectManager *)UserCxt;
	[me toggleTrackingButton:YES];
	printf("Session started. Please wave (%6.2f,%6.2f,%6.2f)...\n", ptFocusPoint.X, ptFocusPoint.Y, ptFocusPoint.Z);
}
// Callback for session end
void XN_CALLBACK_TYPE SessionEnd(void* UserCxt) {
	KinectManager *me = (KinectManager *)UserCxt;
	[me toggleTrackingButton:NO];
	printf("Session ended. Please perform focus gesture to start session\n");
}
// Callback for wave detection
void XN_CALLBACK_TYPE OnWaveCB(void* cxt) {
	printf("Swipe!\n");
}
// callback for a new position of any hand
void XN_CALLBACK_TYPE OnPointUpdate(const XnVHandPointContext* pContext, void* cxt) {
    KinectManager *me = (KinectManager *)cxt;
    
	[me sendValue:(pContext->ptPosition.X + 1500) / 3 fromControlWithID:0];
	[me sendValue:(pContext->ptPosition.Y + 750) / 2 fromControlWithID:1];
	[me sendValue:(pContext->ptPosition.Z + 750) / 2 fromControlWithID:2];		

	printf("%d: (%f,%f,%f) [%f]\n", pContext->nID, pContext->ptPosition.X, pContext->ptPosition.Y, pContext->ptPosition.Z, pContext->fTime);
}

void XN_CALLBACK_TYPE Gesture_Recognized(xn::GestureGenerator& generator, const XnChar * strGesture, const XnPoint3D * pIDPosition, const XnPoint3D * pEndPosition, void *pCookie) {
    printf("gesture recognized was %s\n", strGesture);
}

void XN_CALLBACK_TYPE GeneralSwipe(XnVDirection eDir, XnFloat fVelocity, XnFloat fAngle, void *pUserCxt) {
    printf("swipe direction = %d\n", eDir);
}
void XN_CALLBACK_TYPE SwipeLeft(XnFloat fVelocity, XnFloat fAngle, void *pUserCxt) {
    printf("LEFT SWIPE\n");
}

@implementation KinectManager

- (void) toggleTrackingButton:(BOOL)on {
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	if(on) {
		[view setColor:[NSColor greenColor]];
		[trackingLabel setStringValue: @"TRACKING"];
	}else{
		[view setColor:[NSColor redColor]];
		[trackingLabel setStringValue: @"NOT TRACKING"];
	}
	[pool drain];
}

- (id) init {
	if(self = [super init]) {}
	return self;
}

- (void) awakeFromNib {	
	ds = [NSApp delegate];
	[self createDeviceWithName:@"Kinect"];

	count = 0;
	
    [NSThread detachNewThreadSelector:@selector(sessionThread) toTarget:self withObject:nil];

}

- (void) sessionThread {
    xn::ScriptNode scriptNode;
    XnStatus rc;
    
    context.Init();
    
    g_GestureGenerator.Create(context);
    g_HandsGenerator.Create(context);
    g_HandsGenerator2.Create(context);
    
    g_DepthGenerator.Create(context);
    context.SetGlobalMirror(true);
    
    // Create the Session Manager
    pSessionGenerator = new XnVSessionManager();
    rc = ((XnVSessionManager*)pSessionGenerator)->Initialize(&context, "Click", NULL);
    if (rc != XN_STATUS_OK)
    {
        printf("Session Manager couldn't initialize: %s\n", xnGetStatusString(rc));
        delete pSessionGenerator;
        return;
    }
    
    XnCallbackHandle h1;
    g_GestureGenerator.RegisterGestureCallbacks(Gesture_Recognized, NULL, NULL, h1);

    // Initialization done. Start generating
    context.StartGeneratingAll();
    
    pSessionGenerator->RegisterSession(self, &SessionStart, &SessionEnd, &SessionProgress);
    
	// init & register wave control
	XnVSwipeDetector wc;
	wc.RegisterSwipe(self, GeneralSwipe);
	wc.RegisterPointUpdate(self, OnPointUpdate);
    wc.SetUseSteady(true);
	pSessionGenerator->AddListener(&wc);
    
    while (1) {
        context.WaitAnyUpdateAll();
        ((XnVSessionManager*)pSessionGenerator)->Update(&context);
	}
}

- (void) exampleCallback:(NSTimer *)timer {
	[self sendValue:9999.0f fromControlWithID:0];
}

- (void) createDeviceWithName:(NSString *)name {
	[ds.dm createDeviceWithName:name];
}

- (void) sendValue:(float)value fromControlWithID:(int)controlID {
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	Device *device = [ds.dm deviceForName:@"Kinect"];
	
	if(device.polling && (++count == 30 || count == 31 || count == 32) ) {
		NSString *msg = [[NSString alloc] initWithFormat:@"%@ :: controlID %d :: value %f", device.name, controlID, value];
		[ds.console print:msg];
		[msg release];
		
		if(count == 32) count = 0;
	}
	
	[ds.am sendApplicationsValue:value fromControlWithID:controlID onDevice:device];
	
	[pool drain];
}

- (IBAction) KinectConnect:(id)sender {
	[ds.console print:@"Kinect loaded"];
	[self createDeviceWithName:@"Kinect"];
}

- (NSString *) name { return @"Kinect"; }
- (BOOL) shouldDisplayInterface { return YES; }
- (NSView *) view { return view; }

@end
