#import "PhaseSpaceManager.h"

#define ALLOSPHERE_IP "192.168.0.127"

int _count = 0;
float RIGID_BODY[MARKER_COUNT][3] = {
  { 0, 0,  0},
  { 99.42, 2.15,  -4.34},
  { 90.60, 6.94,  76.32},
};

float _180byPI = 180.0 / M_PI;

void copy_p(const float *a, float *b) { for(int i = 0; i < 7; i++) b[i] = a[i]; }
void print_p(const float *p) { for(int i = 0; i < 7; i++) printf("%f ", p[i]); }

void owl_print_error(const char *s, int n) {
  if(n < 0) printf("%s: %d\n", s, n);
  else if(n == OWL_NO_ERROR) printf("%s: No Error\n", s);
  else if(n == OWL_INVALID_VALUE) printf("%s: Invalid Value\n", s);
  else if(n == OWL_INVALID_ENUM) printf("%s: Invalid Enum\n", s);
  else if(n == OWL_INVALID_OPERATION) printf("%s: Invalid Operation\n", s);
  else printf("%s: 0x%x\n", s, n);
}

@implementation PhaseSpaceManager

- (id) init {
	if(self = [super init]) {}
	return self;
}

- (void) receiveThread {
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	while(1) {
		int err;
		// get the rigid body
		int n = owlGetRigids(&rigid, 1);
		// get markers
		int m = owlGetMarkers(markers, 3);

		// check for error
		if((err = owlGetError()) != OWL_NO_ERROR) {
			owl_print_error("error", err);
			break;
		}
		// no data yet
		if(n == 0) continue;

		if(n > 0) {
			if(rigid.cond > 0) {
				float inv_pose[7];
				copy_p(rigid.pose, inv_pose);
				invert_p(inv_pose);
				if(_count++ %30 == 0) {
					//printf("\nRigid: ");
				  
					//#define POSE(x, y, z, angle, vx, vy, vz) \
					// pose is x, y, z position in mm  w, x, y, z quaternion
				  
					// TODO: TRANSLATE QUATERNION INTO EULER ANGLES, SEND OSC TO DEVICE SERVER
					float qw = rigid.pose[3];
					float qx = rigid.pose[4];
					float qy = rigid.pose[5];
					float qz = rigid.pose[6];
					
					float yaw   = atan2f(2*qy*qw-2*qx*qz , 1 - 2*(qy * qy) - 2*(qz * qz));
					float pitch = asinf(2*qx*qy + 2*qz*qw); 
					float roll  = atan2f(2*qx*qw-2*qy*qz , 1 - 2*(qx * qx) - 2*(qz * qz));
					
					[self sendValue:rigid.pose[0] fromControlWithID:0];
					[self sendValue:rigid.pose[1] fromControlWithID:1];
					[self sendValue:rigid.pose[2] fromControlWithID:2];
					
					[self sendValue:(pitch	* _180byPI)	fromControlWithID:3];
					[self sendValue:(roll	* _180byPI)	fromControlWithID:4];
					[self sendValue:(yaw	* _180byPI)	fromControlWithID:5];
																														
					//printf("yaw = %f, pitch = %f, roll = %f\n", yaw, pitch, roll);
				}
			}
		}
    }
	[pool drain];

	owlDone();
}

- (void) awakeFromNib {
	[PhaseSpaceField1 setStringValue:@"Test String 1"];
	[PhaseSpaceField2 setStringValue:@"PhaseSpace"];
	ds = [NSApp delegate];
}

- (void) exampleCallback:(NSTimer *)timer {
	[self sendValue:9999.0f fromControlWithID:0];
}

- (void) createDeviceWithName:(NSString *)name {
	[ds.dm createDeviceWithName:name];
}

- (void) sendValue:(float)value fromControlWithID:(int)controlID {
	Device *device = [ds.dm deviceForName:@"PhaseSpace"];
	
	if(device.polling)
		[ds.console print:[NSString stringWithFormat:@"%@ :: controlID %d :: value %f", device.name, controlID, value]];
	
	[ds.am sendApplicationsValue:value fromControlWithID:controlID onDevice:device];
}

- (IBAction) PhaseSpaceConnect:(id)sender {
	[ds.console print:@"PhaseSpace loaded"];
	[self createDeviceWithName:@"PhaseSpace"];
	
	int ret = owl_scan(ALLOSPHERE_IP, 8999, "a", 10, &cons, 1);
	if(ret < 0) {
		NSLog(@"ERROR FINDING SERVER");
		return;
	}else{
		NSLog(@"found a server!!!");
		//int err = owlInit(ALLOSPHERE_IP, 0);
	}
	
	int tracker = 0;  
	owlTrackeri(tracker, OWL_CREATE, OWL_RIGID_TRACKER);

	for(int i = 0; i < MARKER_COUNT; i++) {
		owlMarkeri(MARKER(tracker, i), OWL_SET_LED, i);
		owlMarkerfv(MARKER(tracker, i), OWL_SET_POSITION, RIGID_BODY[i]);
	}

	// activate tracker
	owlTracker(tracker, OWL_ENABLE);
	owlSetFloat(OWL_FREQUENCY, 30);
  
	// start streaming
	owlSetInteger(OWL_STREAMING, OWL_ENABLE);
	
	[NSThread detachNewThreadSelector:@selector(receiveThread) toTarget:self withObject:nil];
}

- (NSString *) name { return @"PhaseSpace"; }
- (BOOL) shouldDisplayInterface { return YES; }
- (NSView *) view { return view; }

@end
